
#
# Shiny App for nonlinear support vector separation of Pareto optima
# ******************************************************************
#
# Server File
#
# Press "Run App" in RStudio to start the App
#



library(shiny)
library(rPref)
library(ggplot2)
library(rhandsontable)
library(kernlab)


N <- 20 # fixed number of tuples
pref <- high(x) * high(y) # fixed preference


# Kernels
# =======

# Phi functions of kernels having degree 1 to 5
# ordered by degree of of polynomial w.r.t. x1

phis <- list(
  function(x1, x2) cbind(x1, x2),
  function(x1, x2) cbind(x1^2, sqrt(2) * x1 * x2, x2^2),
  function(x1, x2) cbind(x1^3, sqrt(3) * x1^2 * x2, sqrt(3) * x1 * x2^2, x2^3),
  function(x1, x2) cbind(x1^4, sqrt(4) * x1^3 * x2, sqrt(6) * x1^2 * x2^2, sqrt(4) * x1 * x2^3, x2^4),
  function(x1, x2) cbind(x1^5, sqrt(5) * x1^4 * x2, sqrt(10) * x1^3 * x2^2, 
                         sqrt(10) * x1^2 * x2^3, sqrt(5) * x1 * x2^4, x2^5)
)


generate_mathml <- function(phi, w, b, show_numbers) {
  
  # Formula <w,Phi(x,y)> + b = 0
  formula_main <- c(
    '<mo rspace="0.1em" lspace="0em" form="prefix">&#x0003C;</mo><mi>w</mi><mo>,</mo>
     <mi>&#x003A6;</mi><mrow><mo form="prefix">(</mo><mi>x</mi><mo>,</mo><mi>y</mi>
     <mo form="postfix">)</mo></mrow>
     <mo rspace="0em" lspace="0.1em" form="postfix">&#x0003E;</mo><mo>+</mo><mi>b</mi>',
    '<mo>=</mo>',
    '<mn>0</mn>')
  
  rr <- function(x, pattern, replacement) gsub(pattern, replacement, x)
  rf <- function(x, pattern, replacement) gsub(pattern, replacement, x, fixed = TRUE)
  
  # Base converter function
  base_convert <- function(x) {
    x %>%
      rr("([a-z0-9]+)\\^([0-9]+)", "<msup>\\1 <mn>\\2</mn></msup>") %>%
      rr("sqrt\\(([0-9]+)\\)", "<msqrt><mn>\\1</mn></msqrt>") %>%
      rf("(", '<mo form="prefix">(</mo>') %>%
      rf(")", '<mo form="postfix">)</mo>') %>%
      rf(",", '<mo rspace="0.3em">,</mo>') %>%
      rf("*", '<mo>&#x022C5;</mo>')
  }
  
  # Convert R number into "<mn>...</mn>" string
  extract_number <- function(x) {
    xstr <- format(x, digits = 4)
    if (grepl("e", xstr, fixed = TRUE)) { # Scientific format
      return(rr(xstr, "(\\-?[0-9\\.]+)e\\+([0-9]+)", 
                '<mn>\\1</mn> <mo lspace="0.05em" rspace="0.05em">&#x022C5;</mo> <msup><mn>10</mn> <mn>\\2</mn></msup>'))
    } else {
      return(paste0("<mn>", xstr, "</mn>"))
    }
  }
  
  # Generate w vector
  w_parts <- vapply(w, extract_number, "")
  formula_w <- c('<mi>w</mi>',
                 '<mo>=</mo>',
                 paste0('<mo form="prefix">(</mo>', paste(w_parts, collapse = '<mo rspace="0.3em">,</mo>'),
                        '<mo form="postfix">)</mo>'))
  
  
  rhs_phi <- paste0(deparse(phi)[-1], collapse="") %>%
    rf("cbind", "") %>% 
    base_convert() %>%
    rf("x1", '<mi>x</mi>') %>% rf("x2", '<mi>y</mi>')
    
  formula_phi <- c('<mi>&#x003A6;</mi><mo form="prefix">(</mo><mi>x</mi><mo>,</mo><mi>y</mi><mo form="postfix">)</mo>',
                   '<mo>=</mo>',
                   rhs_phi)
  
  formula_b <- c('<mi>b</mi>',
                 '<mo>=</mo>',
                 extract_number(b))
  
  if (show_numbers) {
    lst_formula <- list(formula_main, formula_phi, formula_w, formula_b)
  } else {
    lst_formula <- list(formula_main, formula_phi)
  }
  
  rows <- vapply(lst_formula, function(x) {
    paste0('<mtr><mtd columnalign="right">',   x[1],
           '</mtd><mtd columnalign="center">', x[2], 
           '</mtd><mtd columnalign="left">',   x[3],
           '</mtd></mtr>')
  }, '')
  
  return(paste0('<math><mtable>', paste0(rows, collapse = ""), '</mtable></math>'))
}


# Shiny Server
# ============

shinyServer(function(input, output, session) {
  
  calculate_levels <- function(df) {
    # Calculate levels
    df <- psel(df, pref, top = nrow(df))
    # Order by row.names (= ids) - note that psel orders by level
    df <- df[order(as.numeric(row.names(df))),]
    
    # Update "Separation level" slider
    updateSliderInput(session, "level_sep", max = max(df$.level))
    
    # Return df with levels
    return(df)
  }
  
  # generate new random values (not reactive!)
  generate_random_vals <- function(cor) {
    
    # Data generator for 2-dim cor/anticor data
    gen_data <- function(N, cor) {  
      rndvals <- matrix(runif(2 * N), N, 2)
      corvals <- matrix(runif(N), N, 2)
      if (cor < 0) { 
        corvals[,2] <- 1 - corvals[,1]
      }
      data <- as.data.frame((1 - abs(cor)) * rndvals + abs(cor) * corvals)
      names(data) <- c("x","y")
      return(data)
    }
    
    # Create data set, add levels and return
    return(calculate_levels(gen_data(N, cor)))
  }
  
  values <- reactiveValues(df = generate_random_vals(isolate(input$cor)))
  
  # this is called when input$recalc is clicked
  observeEvent(input$recalc, {
    # does not react when cor-value changes, just when recalc is clicked
    cor <- input$cor
    # recalc values and store in reactive values
    values$df <- generate_random_vals(cor)
  })
  
  # Update table output
  output$tbl_tuples <- renderRHandsontable({
    rhandsontable(values$df, selectCallback = TRUE, readOnly = FALSE) %>% 
      hot_col(col = 3, readOnly = TRUE) %>% hot_cols(colWidths = c(100,100,50))
  })

  # Write table input to reactive values
  observe({
    if(!is.null(input$tbl_tuples))
      values$df <- calculate_levels(hot_to_r(input$tbl_tuples)[,c('x','y')])
  })
  
  get_svm_dataset <- reactive({
    
    # get data set from generator or table
    df <- values$df
    
    # Level to separate
    level_sep <- input$level_sep
    
    # Calculate Pareto-Optima and Level-n points
    df_cl <- rbind(merge(data.frame(cl =  1), subset(df, .level == 1)),
                   merge(data.frame(cl = -1), subset(df, .level == level_sep)))
    
    # Tuples between level 1 (optima) and level_sep - empty data set for level_sep=2
    df_between <- subset(df, 1 < .level & .level < level_sep)[,c('x', 'y')]
    
    # Tuples lower then level_sep
    df_remainder <- subset(df, level_sep < .level)[,c('x', 'y')]
    
    return(list(df_all = df[,c('x', 'y')], df_cl = df_cl, df_between = df_between, df_remainder = df_remainder,
                df_opt = subset(df_cl, cl == 1 )[,c('x', 'y')],
                df_sec_opt = subset(df_cl, cl == -1)[,c('x', 'y')]))
  })
  
  # Do the separation algorithm and generate the text output
  get_separation <- reactive({
    
    # Initialization
    dfs <- get_svm_dataset()
    df_cl <- dfs$df_cl
    degree <- input$degree
    
    # Check for availability of both classes
    if (sum(df_cl$cl == -1) == 0 || sum(df_cl$cl == 1) == 0) {
      output$state1 <- renderText({ 
        '<span style="color:red"><strong>Just one class, no separation possible</strong></span>' 
      })
      return(list(separation_successful = FALSE))
    }
    
    k  <- function(x, x_) (x %*% t(x_))^degree 
    phi <- phis[[degree]]
     
    # do separation
    # =============
    
    # extract SVM input data from dfs
    x <- as.matrix(df_cl[,c('x','y')])
    y <- df_cl$cl
    
    # ipop solves the quadratic programming problem :
    # \min(c'*x + 1/2 * x' * H * x)
    # subject to: 
    # b <= A * x <= b + r
    # l <= x <= u
    
    eps <- 1e-11    # perturbance size (avoid unstability for non-separable problem)
    n <- nrow(x)  # Number of points
    
    # quadratic problem variables
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    c_vec <- -rep(1, n) # the minus is missing in solve.QP! (in contrast to quadprog)
    
    # H maxtrix using the kernel trick
    H_mat <- (y %*% t(y)) * k(x, x) + eps * diag(n) # add perturbance
    
    # inequality boundary variables
    l_vec <- rep(0, n) # ensure alpha_i >= 0
    u_vec <- rep(1e10, n)  # huge upper bound (quasi-unconstrained problem)
    
    # equality boundary variables (these are scalars!)
    b_scal <- 0
    A_mat <- t(y) # for sum_i (alpha_i y_i) = 0
    r_scal <- 0
    
    # try to solve the optimization problem
    # simultanously increase significant digits and margin exponent
    # defaults are sigf = 7 and margin = 0.05; for i=1 we have sigf = 6 and margin = 0.1
    
    convergence_exp <- 8 # Maximum
    
    for(i in 0:8) {
      margin <- 10^(-i) # margin to boundary
      sigf <- 5+i # significant numbers
      tres <- tryCatch({ipop(c_vec, H_mat, A_mat, b_scal, l_vec, u_vec, r_scal, verb=FALSE, 
                             sigf=sigf, margin=margin)},
                       error = function(e) {i})
      if (is.numeric(tres) > 0) {
        convergence_exp <- tres - 1
        break
      } else {
        res <- tres
      }
    }
    if (convergence_exp == -1) stop("Could not solve the quadratic program!")
    
    alpha <- primal(res)
    
    # SV vector indices are those with non-zero indices
    sv_inds <- (alpha > 1e-5)
    
    # w = sum_i alpha_i * y_i * x_i
    w <- as.vector((alpha[sv_inds] * y[sv_inds]) %*% phi(x[sv_inds,1], x[sv_inds,2]))
    
    # b: averaging over
    #   y_i - k(w, x_i)
    # for all support vectors
    
    bvals <- vapply(which(sv_inds), function(i) (y[i] - ( phi(x[i,1], x[i,2]) %*% w ) ), 0)
    b <- mean(bvals)
    
    # bvals are all nearly the same, if converged
    b_diff <- max(abs(bvals - b))
    separation_successful <- (b_diff < 0.1)
    
    df_sv <- as.data.frame(x[sv_inds,])
    
    if (separation_successful) {
      # Text Output of results
      output$state1 <- renderText({
        paste0("<strong><span style='color:blue'>Separation successful</span></strong><br />",
               "Number of support vectors: <strong>", sum(sv_inds), "</strong><br />",
               "Maximum difference of b-values: ", format(b_diff, digits = 4)) })
      output$state_formula <- renderText({ generate_mathml(phi, w, b, TRUE) })
      
    } else {
      output$state1 <- renderText({"<strong><span style='color:red'>Separation not possible</span><strong>"})
      output$state_formula <- renderText({ paste0("<p>Separation not possible using: </p>",
                                                  generate_mathml(phi, w, b, FALSE))  })
    }
    
    
    # return result
    return(list(separation_successful = separation_successful, b = b, w = w, phi = phi, df_sv = df_sv))
    
  })
  
  # Generate common base ggplot object for separation plot and tuple editor plot
  get_base_plot <- reactive({
    
    # data input
    dfs <- get_svm_dataset()
  
    # Other input
    level_sep <- input$level_sep
    
    # Start calculation/plotting
    # ==========================
    
    # Mapping creator
    get_mapping <- function(str, m = aes(x = x, y = y)) {
      m$colour <- str
      return(m)
    }

    # Tuples names (for legend)
    between_str   <- { if (level_sep == 3) paste0("Level 2 points")
      else                paste0("Level 2-", (level_sep - 1), " points") }
    remainder_str <- paste0("Level >= ", (level_sep + 1), " points")
    opt_str       <- "Pareto optima"
    sec_opt_str   <- paste0("Level ", level_sep, " points")
    
    # Tuples mappings
    m_between   <- get_mapping(between_str)
    m_remainder <- get_mapping(remainder_str)
    m_opt       <- get_mapping(opt_str)
    m_sec_opt   <- get_mapping(sec_opt_str)
    
    # Tuples legend
    tuples_legend_vals        <- c('blue', 'red',        rgb(0.4,0.4,0.4), rgb(0.8,0.8,0.8))
    names(tuples_legend_vals) <- c(opt_str, sec_opt_str, between_str,      remainder_str)
    lgend_idx                 <- c(TRUE,    FALSE,       FALSE,            FALSE)
    
    # Base plot with tuples and legend
    p <- ggplot() + coord_equal(ratio=1, xlim = c(-0.05, 1.05), ylim = c(-0.05, 1.05)) +
      geom_point(m_opt, dfs$df_opt, cex = 4) + 
      scale_color_manual(values = tuples_legend_vals, name = "Points")
    
    # Class to separate
    if (nrow(dfs$df_sec_opt) > 0) {
      steps_sec_opt <- rbind(dfs$df_sec_opt, data.frame(x = c(max(dfs$df_sec_opt$x),0),
                                                        y = c(0,max(dfs$df_sec_opt$y))) )
      
      p <- p + geom_point(m_sec_opt, dfs$df_sec_opt, cex = 4)
      lgend_idx[2] <- TRUE
    }
    
    # Points between
    if (nrow(dfs$df_between) > 0) {
      p <- p + geom_point(m_between, dfs$df_between, cex = 4) 
      lgend_idx[3] <- TRUE
    }
    
    # Remainder points
    if (nrow(dfs$df_remainder) > 0) {
      p <- p + geom_point(m_remainder, dfs$df_remainder, cex = 4)
      lgend_idx[4] <- TRUE
    }
    
    # return plot objet and mappings (needed later for pareto fronts)
    return(list(p = p, mappings = list(m_opt, m_sec_opt)))
  })
  
  
  
  output$editorPlot <- renderPlot({
    
    # Get data
    dfs <- get_svm_dataset()
    
    # Base plot
    p <- get_base_plot()$p
    
    row_index <- input$tbl_tuples_select$select$r
    
    # Add selected row
    if (!is.null(row_index) && row_index > 0) {
      
      df_row <- dfs$df_all[row_index,]

      # mark point
      m_sel <- aes(x = x, y = y, shape = "Selected tuple") 
      p <- p + geom_point(m_sel, df_row, size = 7)
      
      # legend for points
      crit_legend_vals <- c(22)
      names(crit_legend_vals) <- c("Selected tuple")
      p <- p + scale_shape_manual(values = crit_legend_vals, name = "Selected tuple")
    }
    
    
    # Add numbers for tuple editor
    # ============================
    df_num <- dfs$df_all
    df_num$txt <- row.names(df_num)
    m_num <- aes(x=x+0.03, y=y, label=txt)
    p <- p + geom_text(m_num, df_num, size=3, hjust = 0)
    
    # Show plot
    p
  })
  
  
  
  output$skylinePlot <- renderPlot({
  
    # Get data set input (classes to separate, between, remainder)
    dfs <- get_svm_dataset()
    
    # Get separation result
    sep_result <- get_separation()
    separation_successful <- sep_result$separation_successful
    b <- sep_result$b
    w <- sep_result$w
    phi <- sep_result$phi
    df_sv <- sep_result$df_sv
  
    # Other input
    show_sep <- input$show_sep
    
    # Get base plot
    bp <- get_base_plot()
    p <- bp$p
    
    # Add steps
    
    # Steps (Pareto front)
    steps_opt <- rbind(dfs$df_opt, data.frame(x = c(max(dfs$df_opt$x), -0.05),
                                              y = c(-0.05, max(dfs$df_opt$y))) )
    p <- p + geom_step(bp$mappings[[1]], steps_opt, direction="vh")
    
    if (nrow(dfs$df_sec_opt) > 0) {
      steps_sec_opt <- rbind(dfs$df_sec_opt, data.frame(x = c(max(dfs$df_sec_opt$x), -0.05),
                                                        y = c(-0.05, max(dfs$df_sec_opt$y))) )
      p <- p + geom_step(bp$mappings[[2]], steps_sec_opt, direction="vh")
    }
      
    # Show seperation result
    # ======================
    if (show_sep && separation_successful) { 
      
      msl <- aes(x = x, y = y, z = z, linetype = "Seperation line")
      mcor <- aes(x = x, y = y, z = z, linetype = "Seperating corridor")
      
      min_fun <- function(x, y) ((phi(x, y) %*% w) + b)
      
      # Calculate values for min-fun on unit square
      xvec <- seq(-0.05, 1.05, length=100)
      yvec <- xvec
      dfc <- as.data.frame(merge(xvec, yvec))
      dfc$z <- as.vector(outer(xvec, yvec, min_fun))
      # use contour plot to show the zero-line and the +1/-1 lines (corridor)
      p <- p + stat_contour(msl, dfc, breaks=0) + stat_contour(mcor, dfc, breaks = c(-1, 1))
      
      # legend for errors
      sep_legend_vals <- c("solid", "dashed")
      names(sep_legend_vals) <- c("Seperation line", "Seperating corridor")
      
      # mark support vectors
      m_crit <- aes(x = x, y = y, shape = "Support vectors") 
      p <- p + geom_point(m_crit, df_sv, cex = 6)
      
      # legend for support vectors
      crit_legend_vals <- c(22)
      names(crit_legend_vals) <- c("Support vectors")
      p <- p + scale_shape_manual(values = crit_legend_vals, name = "Support vectors") +
        scale_linetype_manual(values = sep_legend_vals, name = "(Approx.) Separating Line")
    
    }
    
    # Show plot
    p
    
  })
})
